/*
 * The strings
 *
 * $Id: str.c 1.21 Sat, 17 Apr 1999 00:42:42 +0200 crad $
 */
#include "s.h"
#include "heap.h"

SOBJ scm_mkstring(char *str)
{
  SOBJ obj = scm_newcell(SOBJ_T_STRING);

  if (str) {
	SCM_STR_LEN(obj)   = strlen(str);
	SCM_STR_VALUE(obj) = scm_must_strdup(str);
  } else {
	SCM_STR_LEN(obj) = 0;
	SCM_STR_VALUE(obj) = NULL;
  }
  return(obj);
}

SOBJ scm_str_alloc(int len)
{
  SOBJ str = scm_newcell(SOBJ_T_STRING);
  SCM_STR_LEN(str) = len;
  SCM_STR_VALUE(str) = scm_must_alloc(len + 1);
  return(str);
}

int scm_strlen(SOBJ str)
{
  if (!SCM_STRINGP(str)) return(-1);
  return(SCM_STR_LEN(str));
}

SOBJ scm_string_append_char(SOBJ str, SOBJ chr)
{
  int c;
  
  if (str == NULL) 	str = scm_mkstring(NULL);
  
  if (!SCM_STRINGP(str)) 	err("want a string", str);
  
  if (!SCM_INUMP(chr) && !SCM_CHARP(chr)) err("bad int or chr", chr);
  c = SCM_INUMP(chr) ? SCM_INUM(chr) : SCM_CHAR(chr);

  if (SCM_STR_VALUE(str)) {		/* allready allocated */
	SCM_STR_VALUE(str) =
	  scm_must_realloc(SCM_STR_VALUE(str), SCM_STR_LEN(str) + 2);
  } else {
	SCM_STR_VALUE(str) = scm_must_alloc(2);
  }
  SCM_STR_VALUE(str)[SCM_STR_LEN(str)++] = c;
  SCM_STR_VALUE(str)[SCM_STR_LEN(str)] = 0;
  return(str);
}

/************************************************************************
 * String type definition
 ************************************************************************/

void scm_string_sweep(SOBJ str)
{
#ifdef DEBUG
  scm_puts("; removing string at ");  port_putx(SCM_OUTP, str); scm_puts("\n");
#endif
  if (str && SCM_STR_VALUE(str)) {
	scm_free(SCM_STR_VALUE(str));
	SCM_STR_VALUE(str) = NULL;
  }
}

void scm_string_print(SOBJ str, PORT *p)
{
  int i;

  for (i = 0; i < SCM_STR_LEN(str); i++) {
	port_putc(p, SCM_STR_VALUE(str)[i]);
  }
}

void scm_string_write(SOBJ str, PORT *p)
{
  int c, i;
  char buf[32];

  port_putc(p, '\"');
  for (i = 0; i < SCM_STR_LEN(str); i++) {
	c = SCM_STR_VALUE(str)[i];
	if (c < ' ' || c == 127) {
	  switch(c) {
	  case 0: 	 port_puts(p, "\\0");	break;
	  case '\a': port_puts(p, "\\a");	break;
	  case '\b': port_puts(p, "\\b");	break;
	  case '\t': port_puts(p, "\\t");	break;
	  case '\n': port_puts(p, "\\n");	break;
	  case '\v': port_puts(p, "\\v");	break;
	  case '\f': port_puts(p, "\\f");	break;
	  case '\r': port_puts(p, "\\r");	break;
	  default:
		sprintf(buf, "\\%uo", c & 0xff);
		port_puts(p, buf);
	  }
	} else {
	  switch(c) {
	  case '\\': port_puts(p, "\\\\"); 	break;
	  case '\"': port_puts(p, "\\\"");	break;
	  default:   port_putc(p, c);
	  }
	}
  }
  port_putc(p, '\"');
}



/*-- detect valid string starter char */
int scm_string_reconize(PORT *port, int c)
{
  return(c == '\"');
}

/*-- read from SCM_INP */
SOBJ scm_string_parse(PORT *port, int start_char)
{
  int c, n;
  SOBJ str = scm_mkstring("");

  while( (c = port_getc(port)) != EOF) {
	if (c == '\"') 	break;
	  
	if (c == '\\')	{			/* escape sequence */
	  if ((c = port_getc(port)) == EOF) {
		err("EOF while reading string...", NULL);
	  }
	  switch(c) {
	  case 'n':		c = '\n';		break;
	  case 'r':		c = '\r';		break;
	  case 'b':		c = '\b';		break;
	  case 't':		c = '\t';		break;
	  case 'x':
		n = 0;
		while( (c = port_peekc(port)) != EOF && isxdigit(c)) {
		  c = port_getc(port);
		  if (c <= '9') {
			c -= '0';
		  } else {
			c = tolower(c) - ('a' + 10);
		  }
		  n = n*16 + c;
		}
	  default:
		if (isdigit(c) && c < '8') {
		  n = c - '0';
		  
		  while(isdigit(port_peekc(port))) {
			c = port_getc(port);
			n = (8 * n) + (c - '0');
		  }
		}
		
	  /* nothing yet */
	  }
	}
	scm_string_append_char(str, SCM_MKINUM(c));
  }
  return(str);
}

/*-- library */

SOBJ scm_string_compare(SOBJ s1, SOBJ s2)
{
  return( (strcmp(SCM_STR_VALUE(s1), SCM_STR_VALUE(s2)) == 0) ?
		  scm_true : scm_false);
}

SOBJ scm_stringp(SOBJ x)
{
  return(SCM_MKBOOL(SCM_STRINGP(x)));
}


SOBJ scm_make_string2(SOBJ l, SOBJ c)
{
  SOBJ s;
  int len, i, j;
  
  if (!SCM_INUMP(l))				err("make-string: bad integer", l);
  if (c != NULL && !SCM_CHARP(c)) 	err("make-string: bad char", c);

  len = SCM_INUM(l);
  s = scm_str_alloc(len);
  j = (c == NULL) ? 0 : SCM_CHAR(c);
  for (i = 0; i < len; i++)  SCM_STR_VALUE(s)[i] = j;
  SCM_STR_VALUE(s)[len] = 0;	/* null terminated for c compatibility */
  return(s);
}

SOBJ scm_make_string(int nargs, SOBJ *a)
{
  if (nargs < 1 || nargs > 2)	err("make-string: bad number of args", NULL);

  return(scm_make_string2(a[0], (nargs == 2) ? a[1] : NULL));
}
  
SOBJ scm_string(int nargs, SOBJ *a)
{
  int i;
  SOBJ str = scm_str_alloc(nargs);

  for (i = 0; i < nargs; i++) {
	SCM_STR_VALUE(str)[i] = SCM_CHAR(a[i]);
  }
  return(str);
}

SOBJ scm_string_length(SOBJ str)
{
  if (!SCM_STRINGP(str)) 	err("string-length: bad string", str);
  return(SCM_MKINUM(SCM_STR_LEN(str)));
}

SOBJ scm_string_ref(SOBJ str, SOBJ index)
{
  int i;
  
  if (!SCM_STRINGP(str))	err("string-ref: bad string", str);
  if (!SCM_INUMP(index))	err("string-ref: bad index", index);

  i = SCM_INUM(index);
  if (i < 0 || i >= SCM_STR_LEN(str))
	err("string-ref: index out of range", index);

  return(scm_mkchar(SCM_STR_VALUE(str)[i]));
}

SOBJ scm_string_set(SOBJ str, SOBJ index, SOBJ chr)
{
  int i;

  if (!SCM_STRINGP(str))	err("string-set!: bad string", str);
  if (!SCM_INUMP(index))	err("string-set!: bad index", index);
  if (!SCM_CHARP(chr))		err("string-set!: bad char", chr);
  
  i = SCM_INUM(index);
  if (i < 0 || i >= SCM_STR_LEN(str))
	err("string-set!: index out of range", index);

  SCM_STR_VALUE(str)[i] = SCM_CHAR(chr);

  return(scm_undefined);
}

static int scmp(SOBJ s1, SOBJ s2)
{
  char *p, *q, *lp, *lq;

  if (!SCM_STRINGP(s1)) 	err("string-cmp: bad string", s1);
  if (!SCM_STRINGP(s2)) 	err("string-cmp: bad string", s2);

  p = SCM_STR_VALUE(s1);  lp = p + SCM_STR_LEN(s1);
  q = SCM_STR_VALUE(s2);  lq = q + SCM_STR_LEN(s2);
  while(p < lp && q < lq) {
	if (*p - *q != 0) return(*p - *q);
	p++;
	q++;
  }
  if (p == lp && q == lq) 	return(0);
  return(*p - *q);
}

static int scmpi(SOBJ s1, SOBJ s2)
{
  char *p, *q, *lp, *lq;

  if (!SCM_STRINGP(s1)) 	err("string-cmp: bad string", s1);
  if (!SCM_STRINGP(s2)) 	err("string-cmp: bad string", s2);

  p = SCM_STR_VALUE(s1);  lp = p + SCM_STR_LEN(s1);
  q = SCM_STR_VALUE(s2);  lq = q + SCM_STR_LEN(s2);
  while(p < lp && q < lq) {
	if (tolower(*p) - tolower(*q) != 0) return(tolower(*p) - tolower(*q));
	p++;
	q++;
  }
  if (p == lp && q == lq) 	return(0);
  return(tolower(*p) - tolower(*q));
}

SOBJ scm_string_lt(SOBJ s1, SOBJ s2) { return(SCM_MKBOOL(scmp(s1,s2) <  0)); }
SOBJ scm_string_le(SOBJ s1, SOBJ s2) { return(SCM_MKBOOL(scmp(s1,s2) <= 0)); }
SOBJ scm_string_eq(SOBJ s1, SOBJ s2) { return(SCM_MKBOOL(scmp(s1,s2) == 0)); }
SOBJ scm_string_ge(SOBJ s1, SOBJ s2) { return(SCM_MKBOOL(scmp(s1,s2) >= 0)); }
SOBJ scm_string_gt(SOBJ s1, SOBJ s2) { return(SCM_MKBOOL(scmp(s1,s2) >  0)); }

SOBJ scm_string_ci_lt(SOBJ s1, SOBJ s2){return(SCM_MKBOOL(scmpi(s1,s2) <  0));}
SOBJ scm_string_ci_le(SOBJ s1, SOBJ s2){return(SCM_MKBOOL(scmpi(s1,s2) <= 0));}
SOBJ scm_string_ci_eq(SOBJ s1, SOBJ s2){return(SCM_MKBOOL(scmpi(s1,s2) == 0));}
SOBJ scm_string_ci_ge(SOBJ s1, SOBJ s2){return(SCM_MKBOOL(scmpi(s1,s2) >= 0));}
SOBJ scm_string_ci_gt(SOBJ s1, SOBJ s2){return(SCM_MKBOOL(scmpi(s1,s2) >  0));}

SOBJ scm_substring(SOBJ string, SOBJ start, SOBJ end)
{
  SOBJ s;
  int len;
  
  if (!SCM_STRINGP(string)) 	err("substring: bad string", string);
  if (!SCM_INUMP(start))		err("substring: bad start", start);
  if (!SCM_INUMP(end))			err("substring: bad end", start);
  
  len = SCM_INUM(end) - SCM_INUM(start);
  if (len < 0) 					err("substring: start is not <= end", start);

  s = scm_str_alloc(len);
  strncpy(SCM_STR_VALUE(s),
		  SCM_STR_VALUE(string) + SCM_INUM(start),
		  len);
  SCM_STR_VALUE(s)[len] = 0;
  return(s);
}

SOBJ scm_string_append(int nargs, SOBJ *s)
{
  int i, len;
  SOBJ str;
  char *p;
  
  len = 0;
  for (i = 0; i < nargs; i++) {
	if (!SCM_STRINGP(s[i])) 	err("string-append: bad string", s[i]);
	len += SCM_STR_LEN(s[i]);
  }
  str = scm_str_alloc(len);
  p = SCM_STR_VALUE(str);
  for (i = 0; i < nargs; i++) {
	strncpy(p, SCM_STR_VALUE(s[i]), SCM_STR_LEN(s[i]));
	p += SCM_STR_LEN(s[i]);
  }
  return(str);
}

/*S* (string-append2 str1 str2) */
SOBJ scm_string_append2(SOBJ str1, SOBJ str2)
{
  SOBJ str;
  if (!SCM_STRINGP(str1)) err("string-append2: bad string", str1);
  if (!SCM_STRINGP(str2)) err("string-append2: bad string", str2);

  str = scm_str_alloc(SCM_STR_LEN(str1) + SCM_STR_LEN(str2));
  strcpy(SCM_STR_VALUE(str), SCM_STR_VALUE(str1));
  strcpy(SCM_STR_VALUE(str) + SCM_STR_LEN(str1), SCM_STR_VALUE(str2));
  return(str);
}

/*-- string coerse */

SOBJ scm_string_to_list(SOBJ str)
{
  SOBJ list;
  int i;
  
  if (!SCM_STRINGP(str))	err("string->list: bad string", str);

  list = NULL;
  for (i = SCM_STR_LEN(str)-1; i >= 0; --i) {
	list = scm_cons(scm_mkchar(SCM_STR_VALUE(str)[i]), list);
  }
  return(list);
}

/*-- ??? make an intern symbol ? */
SOBJ scm_string_to_symbol(SOBJ str)
{
  if (!SCM_STRINGP(str))	err("string->symbol: bad string", str);
  return(scm_mksymbol(SCM_STR_VALUE(str)));
}

SOBJ scm_symbol_to_string(SOBJ sym)
{
  switch(SCM_OBJTYPE(sym)) {
  case SOBJ_T_SYMBOL:  return(scm_atom_to_string(SCM_SYM_NAME(sym)));
  case SOBJ_T_ATOM:	   return(scm_atom_to_string(sym));
  }
  err("symbol->string: bad symbol", sym);
  return(scm_undefined);
}


SOBJ scm_list_to_string(SOBJ list)
{
  SOBJ str;
  int len;
  char *p;

  len = scm_list_length(list);
  if (len < 0) err("list->string: bad list", list);
  
  str = scm_str_alloc(len);
  p = SCM_STR_VALUE(str);
  while(list) {
	if (!SCM_CHARP(SCM_CAR(list)))
	  err("list->string: bad char", SCM_CAR(list));
	*p++ = SCM_CHAR(SCM_CAR(list));
	list = SCM_CDR(list);
  }
  *p = 0;
  return(str);
}

SOBJ scm_string_copy(SOBJ str)
{
  SOBJ s;
  
  if (!SCM_STRINGP(str))	err("string-copy: bad string", str);

  s = scm_str_alloc(SCM_STR_LEN(str));
  strncpy(SCM_STR_VALUE(s), SCM_STR_VALUE(str), SCM_STR_LEN(str));
  return(s);
}

/*S* (string-fill STR CHR) */
SOBJ scm_string_fill(SOBJ str, SOBJ chr)
{
  int i;
  if (!SCM_STRINGP(str))	err("string-fill: bad string", str);
  if (!SCM_CHARP(chr))		err("string-fill: bad char", chr);
  
  for (i = 0; i < SCM_STR_LEN(str); i++) {
	SCM_STR_VALUE(str)[i] = SCM_CHAR(chr);
  }
  return(scm_undefined);
}

/*E* (string-index SEARCHED STRING) => index | #f */
SOBJ scm_string_index(SOBJ ssearch, SOBJ instr)
{
  char *is, *ss, *q;
  int is_len, ss_len;

  if (!SCM_STRINGP(ssearch))  err("string-index: bad search string", ssearch);
  if (!SCM_STRINGP(instr))	  err("string-index: bad string", instr);

  ss = SCM_STR_VALUE(ssearch);  ss_len = SCM_STR_LEN(ssearch);
  is = SCM_STR_VALUE(instr);   	is_len = SCM_STR_LEN(instr);
  q = is + (is_len - ss_len);
  
  while( is <= q ) {
	if (memcmp(is, ss, ss_len) == 0)
	  return(SCM_MKINUM((is - SCM_STR_VALUE(instr))));
	is++;
  }
  return(scm_false);
}


/*E* (string-chop OBJ) => OBJ */
SOBJ scm_string_chop(SOBJ str)
{
  char *p, *l;
  
  if (!SCM_STRINGP(str)) return(str);
  
  p = SCM_STR_VALUE(str);
  l = p + SCM_STR_LEN(str);
  while(p < l) {
	if (*p == '\n') {
	  *p = 0;
	  SCM_STR_LEN(str) = p - SCM_STR_VALUE(str);
	  break;
	}
	p++;
  }
  return(str);
}

/*E* (string-split DELIM STRING) => <list> */
SOBJ scm_string_split(SOBJ delim, SOBJ str)
{
  char *p, *l, *s;
  SOBJ res;
  
  if (!SCM_STRINGP(str))	err("bad string", str);
  if (!SCM_STRINGP(delim))	err("bad string", delim);

  p = SCM_STR_VALUE(str);
  l = p + SCM_STR_LEN(str);
  s = p;
  res = NULL;
  
  while(p < l) {
	if (strchr(SCM_STR_VALUE(delim), *p)) { /* is a delim ? */
	  SOBJ tmp = scm_str_alloc( p - s );
	  strncpy(SCM_STR_VALUE(tmp), s, p - s);
	  SCM_STR_VALUE(tmp)[p-s] = 0;
	  res = scm_cons(tmp, res);
	  s = p+1;
	}
	p++;
  }
  res = scm_cons(scm_mkstring(s), res);
  return(scm_reverse(res));
}

/*E* (string-join SEP LIST) => STRING */
SOBJ scm_string_join(SOBJ sep, SOBJ list)
{
  SOBJ l;
  int len, seplen, nstring;
  char *p;
  SOBJ str;

  if (!SCM_STRINGP(sep)) 		err("bad string", sep);
  len = 0;
  seplen = SCM_STR_LEN(sep);
  nstring = 0;
  for (l = list; l; l = SCM_CDR(l)) {
	if (!SCM_PAIRP(l))				err("bad list", list);
	if (!SCM_STRINGP(SCM_CAR(l)))	err("bad string", SCM_CAR(l));
	len += SCM_STR_LEN(SCM_CAR(l));
	nstring++;
  }
  if (nstring > 0) {
	str = scm_str_alloc(len + ((nstring - 1) * seplen));
	p = SCM_STR_VALUE(str);
	for (l = list; l; l = SCM_CDR(l)) {
	  strncpy(p, SCM_STR_VALUE(SCM_CAR(l)), SCM_STR_LEN(SCM_CAR(l)));
	  p += SCM_STR_LEN(SCM_CAR(l));
	  if (SCM_CDR(l)) {
		strncpy(p, SCM_STR_VALUE(sep), seplen);
		p += seplen;
	  }
	}
	return(str);
  }
  return(scm_mkstring(""));
}

/*E* (string-lower STRING) => STRING */
SOBJ scm_string_lower(SOBJ str)
{
  SOBJ new;
  char *p, *l, *d;
  
  if (!SCM_STRINGP(str)) err("string-lower: bad string", str);
  
  new = scm_str_alloc(SCM_STR_LEN(str));
  
  d = SCM_STR_VALUE(new);
  p = SCM_STR_VALUE(str);
  l = p + SCM_STR_LEN(str);
  while(p < l) {
	*d = tolower(*p);
	p++;
	d++;
  }
  *d = 0;
  return(new);
}

/*E* (string-upper STRING) => STRING */
SOBJ scm_string_upper(SOBJ str)
{
  SOBJ new;
  char *p, *l, *d;
  
  if (!SCM_STRINGP(str)) err("string-upper: bad string", str);
  
  new = scm_str_alloc(SCM_STR_LEN(str));
  
  d = SCM_STR_VALUE(new);
  p = SCM_STR_VALUE(str);
  l = p + SCM_STR_LEN(str);
  while(p < l) {
	*d = toupper(*p);
	p++;
	d++;
  }
  *d = 0;
  return(new);
}


void scm_init_str()
{
  /*-- r5rs string */
  scm_add_cprim("string?",			scm_stringp,		1);
  scm_add_cprim("make-string",		scm_make_string,	-1);
  scm_add_cprim("make-string2",		scm_make_string2,	2);
  scm_add_cprim("string",			scm_string,			-1);
  scm_add_cprim("string-length",	scm_string_length,	1);
  scm_add_cprim("string-ref",		scm_string_ref,		2);
  scm_add_cprim("string-set!",		scm_string_set,		3);

  scm_add_cprim("string<?",			scm_string_lt,		2);
  scm_add_cprim("string<=?",		scm_string_le,		2);
  scm_add_cprim("string=?",			scm_string_eq,		2);
  scm_add_cprim("string>=?",		scm_string_ge,		2);
  scm_add_cprim("string>?",			scm_string_gt,		2);

  scm_add_cprim("string-ci<?",		scm_string_ci_lt,		2);
  scm_add_cprim("string-ci<=?",		scm_string_ci_le,		2);
  scm_add_cprim("string-ci=?",		scm_string_ci_eq,		2);
  scm_add_cprim("string-ci>=?",		scm_string_ci_ge,		2);
  scm_add_cprim("string-ci>?",		scm_string_ci_gt,		2);

  scm_add_cprim("substring",		scm_substring,			3);
  scm_add_cprim("string-append",	scm_string_append, 		-1);

  scm_add_cprim("string->list",		scm_string_to_list,		1);
  scm_add_cprim("list->string",		scm_list_to_string,		1);
  scm_add_cprim("string->symbol",	scm_string_to_symbol,	1);
  scm_add_cprim("symbol->string",  	scm_symbol_to_string,	1);
  scm_add_cprim("string-copy",		scm_string_copy,		1);
  scm_add_cprim("string-fill!",		scm_string_fill,		2);
  
  /*-- string extensions */

  scm_add_cprim("string-append2",		scm_string_append2, 	2);
  scm_add_cprim("string-append-char", 	scm_string_append_char, 2);

  scm_add_cprim("string-index",		scm_string_index, 	2);
  scm_add_cprim("string-chop",		scm_string_chop,  	1);
  scm_add_cprim("string-split",		scm_string_split, 	2);
  scm_add_cprim("string-join",		scm_string_join,  	2);
  scm_add_cprim("string-lower",		scm_string_lower,	1);
  scm_add_cprim("string-upper",		scm_string_upper,	1);
}
