/*
 * Generic variable system: may inteface any type of Variable
 *
 * $Id$
 */
#include "s.h"

extern SCM_VarAux scm_var_hook[];

void scm_var_mark(SOBJ var)
{
}

void scm_var_sweep(SOBJ var)
{
  SCM_VarAux *p, *aux;

  aux = SCM_VAR_AUX(var);
  if (aux) {
	for (p = scm_var_hook; p->name; p++) {
	  if (p == SCM_VAR_AUX(var)) { /* static data */
		return;
	  }
	}
	scm_free(aux);
	SCM_VAR_AUX(var) = NULL;
  }
}

void scm_var_print(SOBJ x, PORT *p)
{
  port_puts(p, "<#var ");
  port_puts(p, SCM_VAR_AUX(x)->name);
  port_puts(p, ">");
}

void scm_var_write(SOBJ x, PORT *p)
{
  port_puts(p, "<#var ");
  port_puts(p, SCM_VAR_AUX(x)->name);
  port_puts(p, ">");
}

#define FETCH(t,ptr) 		(*((t *)(ptr)))
#define STORE(t,ptr,value) { *((t *)(ptr)) = value; }

#define VFETCH(t,var) 			FETCH(t, SCM_VAR_ADDR(var))
#define VSTORE(t,var,value)		STORE(t, SCM_VAR_ADDR(var), value)

/*-- driver functions for various types of external variables
 *
 */

static SOBJ get_ext_char(SOBJ var) {
  return(scm_mkchar(VFETCH(char, var)));
}

static void set_ext_char(SOBJ var, SOBJ value) {
  if (!SCM_CHARP(value)) err("set_ext_char: bad char", value);
  VSTORE(char, var, SCM_CHAR(value));
}

static SOBJ get_ext_short(SOBJ var) {
  return(SCM_MKINUM(VFETCH(short, var)));
}

static void set_ext_short(SOBJ var, SOBJ value) {
  VSTORE(short, var, scm_number2long(value));
}

static SOBJ get_ext_long(SOBJ var) {
  return(scm_int2num(VFETCH(long, var)));
}

static void set_ext_long(SOBJ var, SOBJ value) {
  VSTORE(long, var, scm_number2long(value));
}

static SOBJ get_ext_float(SOBJ var) {
  return(scm_mkfnum(VFETCH(float, var)));
}

static void set_ext_float(SOBJ var, SOBJ value) {
  VSTORE(float, var, scm_number2double(value));
}

static SOBJ get_ext_double(SOBJ var) {
  return(scm_mkfnum(VFETCH(double, var)));
}

static void set_ext_double(SOBJ var, SOBJ value) {
  VSTORE(double, var, scm_number2double(value));
}

static SOBJ get_ext_str(SOBJ var) {
  return(scm_mkstring(VFETCH(char *, var)));
}

static void set_ext_str(SOBJ var, SOBJ value) {

  if (!SCM_STRINGP(value)) 	err("bad string", value);
  if (VFETCH(char *, var) != NULL) {
	free(VFETCH(char *, var));
  }
  VSTORE(char *, var, strdup(SCM_STR_VALUE(value)));
}

static SOBJ get_ext_strbuf(SOBJ var) {
  return(scm_mkstring(SCM_VAR_ADDR(var)));
}

static void set_ext_strbuf(SOBJ var, SOBJ value) {
  if (!SCM_STRINGP(value)) 	err("bad string", value);
  strcpy((char*)SCM_VAR_ADDR(var), SCM_STR_VALUE(value));
}

static SOBJ get_cscheme(SOBJ var) {
  return(VFETCH(SOBJ, var));
}

static void set_cscheme(SOBJ var, SOBJ value) {
  VSTORE(SOBJ , var, value);
}

SCM_VarAux scm_var_hook[] = {
  { ":char",	NULL,	get_ext_char, NULL, 	set_ext_char, NULL, 	0 },
  { ":short", 	NULL,	get_ext_short, NULL, 	set_ext_short, NULL, 	0 },
  { ":long", 	NULL,	get_ext_long, NULL, 	set_ext_long, NULL,		0 },
  { ":float", 	NULL,	get_ext_float, NULL, 	set_ext_float, NULL, 	0 },
  { ":double",	NULL,	get_ext_double, NULL, 	set_ext_double, NULL, 	0 },
  { ":string",  NULL,	get_ext_str, NULL,		set_ext_str,  NULL,		0 },
  { ":string-buffer", NULL, get_ext_strbuf, NULL, set_ext_strbuf, NULL, 0 },
  { ":cscheme",	NULL,	get_cscheme, NULL, 		set_cscheme, NULL, 		0 },
  
  { NULL }
};

static SCM_VarAux *scm_var_type_lookup(SOBJ atom)
{
  SCM_VarAux *va = scm_var_hook;
  while(va->name) {
	if (va->atom == atom) return(va);
	va++;
  }
  return(NULL);
}

static void scm_var_type_init()
{
  SCM_VarAux *va = scm_var_hook;
  while(va->name) {
	va->atom = scm_mkatom(va->name);
	va++;
  }
}

SOBJ scm_mkvar(SOBJ type, void *addr)
{
  SOBJ var;
  SCM_VarAux *aux;

  aux = scm_var_type_lookup(type);
  if (aux == NULL) 	err("mkvar: bad type", type);
  var = scm_newcell(SOBJ_T_VAR);
  SCM_VAR_AUX(var) = aux;
  SCM_VAR_ADDR(var) = addr;
  return(var);
}

SOBJ scm_make_extern_var(SOBJ lib, SOBJ type, SOBJ name)
{
  SOBJ var;
  void *addr;

  if (!SCM_STRINGP(lib))   err("make-extern-var: bad lib",  lib);
  if (!SCM_KEYWORDP(type)) err("make-extern-var: bad type", type);
  if (!SCM_STRINGP(name))  err("make-extern-var: bad name", name);
  
  addr = scm_find_extsym(SCM_STR_VALUE(lib), SCM_STR_VALUE(name), TRUE);
  var = scm_mkvar(SCM_KEYW_NAME(type), addr);
  return(var);
}

void scm_add_cvar(char *name, SOBJ *addr)
{
  SOBJ var = scm_mkvar(scm_mkatom(":cscheme"), addr);
  scm_symadd(name, var);
  scm_gc_protect(addr);
}

void scm_init_variable()
{
  scm_var_type_init();
  scm_add_cprim("make-extern-variable",	&scm_make_extern_var, 3);

  /* init system variable. it helps for initialisation sequence */
  scm_add_cvar("atom-hash", 	&scm_atom_hash);
  scm_add_cvar("symbol-hash", 	&scm_symbol_hash);
  scm_add_cvar("keyword-hash", 	&scm_keyword_hash);

  scm_add_cvar("stdin-port",	&scm_in_port);
  scm_add_cvar("stdout-port",	&scm_out_port);
  scm_add_cvar("stderr-port",	&scm_err_port);
}
